/**
 * Graphics module for Photon.  Comments have been added to the relevant sections.
 * DO NOT MODIFY THIS FILE.
 */

#include "photon-graphics.h"
#include "genlib.h"
#include "extgraph.h"
#include "sound.h"
#include <cstdlib>
#include <stdexcept>
#include <sstream>
#include <map>
#include <string>
#include <iostream>
using namespace std;

/**** MODULE CONSTANTS ****/

static const int TILE_SIZE = 25;                    // Pixels per tile
static const char *const SWAP_SOUND = "drop.wav";   // Sound file for swapping tiles
static const char *const WIN_SOUND  = "win.wav";    // Sound file for victory
static const char *const SELECT_SOUND = "select.wav"; // Sound file for selecting first tile.

/**** MODULE GLOBALS ****/

/* The auto_ptr containing the single PhotonGraphics instance. */
auto_ptr<PhotonGraphics> PhotonGraphics::msInstance;

/**** FUNCTION PROTOTYPES ****/
static void HighlightLocation(int x, int y, double startX, double startY);

/**** MODULE IMPLEMENTATION ****/

/* Returns the stored instance, creating one if it doesn't already exist. */
PhotonGraphics* PhotonGraphics::getInstance()
{
	if(msInstance.get() != NULL)
		return msInstance.get();

	msInstance.reset(new PhotonGraphics);
	return msInstance.get();
}

/* Constructor sets everything to zeros. */
PhotonGraphics::PhotonGraphics() : mWidth(0), mHeight(0), mStartX(0.0),
	mStartY(0.0), mIsRunning(false), mController(NULL)
{
}

/* setController works as advertised. */
void PhotonGraphics::setController(PhotonController* aController)
{
	if(mIsRunning)
		throw runtime_error("Cannot call setController while game is running.");
	
	if(!aController)
		throw invalid_argument("NULL parameter to setController.");

	mController = aController;
}

/* Resizes the world and computes graphics locations. */
void PhotonGraphics::resizeWorld(int aWidth, int aHeight)
{
	if(mIsRunning)
		throw runtime_error("Cannot call resizeWorld while game is running.");

	if(aWidth <= 0 || aHeight <= 0)
		throw invalid_argument("Arguments to resizeWorld must be positive.");

	if(aWidth >= MAX_DIMENSIONS || aHeight >= MAX_DIMENSIONS)
		throw invalid_argument("Arguments to resizeWorld must be less than MAX_DIMENSIONS.");

	mWidth = aWidth;
	mHeight = aHeight;

	/* Now, compute the start X and Y. */
	const double numPixels = TILE_SIZE * MAX_DIMENSIONS;
	mStartX = ((numPixels - aWidth  * TILE_SIZE) / GetXResolution()) / 2;
	mStartY = ((numPixels - aHeight * TILE_SIZE) / GetYResolution()) / 2;
}

/* Helper functions that converts between grid coordinates and graphics coordinates. */
static inline double GetXFor(int x, double startX)
{
	return startX + (x * TILE_SIZE) / GetXResolution();
}
static inline double GetYFor(int y, double startY)
{
	return startY + (y * TILE_SIZE) / GetYResolution();
}

/* Draws a single tile to the screen at the specified location. */
static void PlotTile(tileT aTile, int aX, int aY, double startX, double startY)
{
	/* We cache the filenames associated with tileTs that we've plotted before. */
	static map<tileT, string> kFilenameMap;

	MovePen(GetXFor(aX, startX), GetYFor(aY, startY));

	/* Memoize the filename. */
	if(!kFilenameMap.count(aTile))
	{
		stringstream converter;
		converter << aTile << ".bmp";
		kFilenameMap[aTile] = converter.str();
	}

	DrawNamedPicture(kFilenameMap[aTile]);
}

/* Fills the window with a black rectangle. */
static void FillWindow()
{
	SetPenColor("Black");
	MovePen(0, 0);
	StartFilledRegion(1);
	DrawLine(GetWindowWidth(), 0);
	DrawLine(0, GetWindowHeight());
	DrawLine(-GetWindowWidth(), 0);
	DrawLine(0, -GetWindowHeight());
	EndFilledRegion();
}

/* Enters the main game loop. */
void PhotonGraphics::runGame()
{
	if(!mController)
		throw runtime_error("Must have a controller to process input.");
	if(mIsRunning)
		throw runtime_error("Cannot call processInput when the game is running.");
	if(mWidth == 0 || mHeight == 0)
		throw runtime_error("Must set dimensions of world before calling processInput");

	/* Set up graphics. */
	const double numPixels = TILE_SIZE * MAX_DIMENSIONS;
	SetWindowSize(numPixels / GetXResolution(), numPixels / GetYResolution());
	InitGraphics();
	SetWindowTitle("Photon");

	/* Make the background black for effect. */
	FillWindow();

	/* Start the game! */
	mIsRunning = true;

	/* Tell the controller we're starting. */
	mController->onGameStart();

	/* Now, sit in a loop continuously moving tiles until the game is over. */
	while(!mController->isPuzzleSolved())
	{
		/* Draw the board so that we can see things. */
		mController->drawBoard();

		/* Get two points to swap. */
		pair<int, int> choices[2];
		for(int i = 0; i < 2; ++i)
		{
			/* Keep looping until we get something we like. */
			while(true)
			{
				WaitForMouseDown();
				choices[i].first  = static_cast<int>(((GetMouseX() - mStartX) * GetXResolution()) / TILE_SIZE);
				choices[i].second = static_cast<int>(((GetMouseY() - mStartY) * GetYResolution()) / TILE_SIZE);
				WaitForMouseUp(); 

				/* Now, validate that this is a legitimate point. */
				if(choices[i].first >= 0 && choices[i].second >= 0 &&
				   choices[i].first < mWidth && choices[i].second < mHeight &&
				   mController->trySelectTile(choices[i].first, choices[i].second, i == 0))
					break;

			}
			if(i == 0)
			{
				HighlightLocation(choices[i].first, choices[i].second, mStartX, mStartY);
				PlayNamedSound(SELECT_SOUND);
			}
		}
		PlayNamedSound(SWAP_SOUND);
		mController->swapTiles(choices[0].first, choices[0].second, choices[1].first, choices[1].second);
	}
	
	/* Draw the board so that we can see the victory state. */
	mController->drawBoard();
	mController->onGameEnd();

	PlayNamedSound(WIN_SOUND);
	mIsRunning = false;
}

void PhotonGraphics::drawTile(tileT aTile, int aX, int aY, const set<directionT> &aLasers)
{
	/* Plot the actual tile. */
	PlotTile(aTile, aX, aY, mStartX, mStartY);

	/* For each beam, draw the beam. */
	for(set<directionT>::const_iterator itr = aLasers.begin(); itr != aLasers.end(); ++itr)
	{
		const double originX = (GetXFor(aX, mStartX) + GetXFor(aX + 1, mStartX)) / 2;
		const double originY = (GetYFor(aY, mStartY) + GetYFor(aY + 1, mStartY)) / 2;
		MovePen(originX, originY);
		SetPenColor("Red");
		
		/* Based on what direction we go, we need to draw a different segment.  The numbers here
		 * include a good deal of fudging to make the beams seem continuous.
		 */
		switch(*itr)
		{
		case West:
			DrawLine((-TILE_SIZE / 2 - 2) / GetXResolution(), 0);
			break;
		case East:
			DrawLine((TILE_SIZE / 2 + 1) / GetXResolution(), 0);
			break;
		case North:
			DrawLine(0, (TILE_SIZE / 2 + 1) / GetYResolution());
			break;
		case South:
			DrawLine(0, (-TILE_SIZE / 2 - 1) / GetYResolution());
			break;
		}
	}

}

/* Draws a circle around the specified location to indicate to the user that they have
 * selected an object.
 */
void HighlightLocation(int aX, int aY, double aStartX, double aStartY)
{
	const double originX = (GetXFor(aX + 2, aStartX) + GetXFor(aX + 1, aStartX)) / 2;
	const double originY = (GetYFor(aY, aStartY) + GetYFor(aY + 1, aStartY)) / 2;
	const double radius  = TILE_SIZE / GetXResolution();
	MovePen(originX, originY);
	SetPenColor("White");
	DrawEllipticalArc(radius, radius, 0, 360);
}