/***************************************
 * File: photon-graphics.h
 *
 * Graphics module for Photon.  PhotonGraphics is responsible for gluing together the user
 * input and graphics functions into one cohesive module.  DO NOT MODIFY THE CONTENTS OF
 * THIS FILE.
 */

#ifndef PhotonGraphics_included_
#define PhotonGraphics_included_

#include "direction.h"
#include "photon-controller.h"
#include "tile.h"
#include <set>
#include <memory>
using namespace std;

/**
 * A class which displays a MyPhoton board and which is responsible for driving user input.
 * PhotonGraphics is a singleton class and you cannot instantiate it directly.  To get a pointer
 * to the single instance of PhotonGraphics, use the getInstance static member function.
 */
class PhotonGraphics
{
public:
	/**
	 * A constant defining the maximum possible world dimensions along any axis.
	 * While theoretically we could have a larger world, the constraints of the
	 * graphics module mean that any worlds we read in must be smaller than this
	 * size.  This restriction is already enforced in the skeletal world reading
	 * file, so you shouldn't have to worry about this.
	 */
	static const int MAX_DIMENSIONS = 20;

	/**
	 * PhotonGraphics is a singleton class, so it is illegal to create
	 * more than one copy.  This function returns the single instance of
	 * the PhotonGraphics object.
	 */
	static PhotonGraphics* getInstance();

	/**
	 * Sets which object will be informed about user input requests.  The caller
	 * is responsible for cleaning up this object.
	 *
	 * @param aController The callback object to associate with this graphics
	 *        object.  If this parameter is NULL, setController throws
	 *        an invalid_argument exception.  If the game is currently running,
	 *        throws a runtime_error exception.
	 */
	void setController(PhotonController* aController);

	/**
	 * Resizes the display to the specified width and height.  If either
	 * parameter is non-positive or greater than the constant MAX_DIMENSION,
	 * throws an invalid_argument exception.  If the game is running, throws a
	 * runtime_error.
	 *
	 * @param aWidth The width to make the display
	 * @param aHeigth The height to make the display.
	 */
	void resizeWorld(int aWidth, int aHeight);

	/**
	 * Draws a game tile.  There are several parameters to this function, which
	 * are documented below.
	 *
	 * @param aTile The tile to draw as the background.
	 * @param aX The X coordinate at which to plot the tile.
	 * @param aY The Y coordinate at which to plot the tile.
	 * @param aLasers A set of directionTs corresponding to where the lasers are at this tile.
	 *        For example, if this is a / mirror with an incoming laser from the west and an
	 *        outgoing laser leaving in the north, the set would contain North and West.
	 * If any of these parameters are invalid, throws an invalid_argument.  If the dimensions
	 * have not been set, throws a runtime_error.
	 */
	void drawTile(tileT aTile, int aX, int aY, const set<directionT> & aLasers);

	/**
	 * Sits in a loop waiting for commands from the user and sending commands to the
	 * controller object.  This function only returns when the controller's
	 * isPuzzleSolved function returns true.
	 *
	 * If the controller hasn't been provided or the world hasn't been sized, this
	 * throws a runtime_error.
	 */
	void runGame();

private:
	PhotonGraphics(); // Cannot construct objects.
	PhotonGraphics(const PhotonGraphics&); // Cannot copy
	PhotonGraphics& operator= (const PhotonGraphics& ); // Cannot copy

	int mWidth, mHeight;
	double mStartX, mStartY;
	PhotonController *mController;
	bool mIsRunning;
	static auto_ptr<PhotonGraphics> msInstance;
};

#endif